/* 
 * FILE --- fuzzy_mf.hh
 *
 * PURPOSE --- Header file for a class for fuzzy membership functions (MF).
 * The MF could be chosen to be either triangular or Gaussian. Every other 
 * MF type is interpreted as triangular MF. The user specifies the center 
 * and the spread of the MF.
 */

// Prevent Multiple Inclusion
#ifndef FUZZY_MF_HH
#define FUZZY_MF_HH

// Include Files
#include <stdlib.h>

// Local Definitions and Macros
#define TRIANGULAR_MF 0   // Triangular MF
#define GAUSSIAN_MF 1   // Gaussian MF

// Useful macros
#define MAX(A,B)  ((A) > (B) ? (A) : (B))
#define MIN(A,B)  ((A) < (B) ? (A) : (B))
#define SATURATE(u,MAX_VALUE) \
((u) > MAX_VALUE ? MAX_VALUE : ((u) < -MAX_VALUE ? -MAX_VALUE : (u)));

class FUZZY_MF
{
public:
  // Default constructor
  FUZZY_MF();

  // Constructor
  FUZZY_MF(double _center, double _spread, int _type = TRIANGULAR_MF);

  // Destructor
  ~FUZZY_MF() {};

  // Sets the given MF to the input MF _mf
  void set_mf(const FUZZY_MF &_mf);

  // Set or return the center of the MF
  void set_center(double _center) { center = _center; }
  double get_center(void) { return center; }

  // Set or return the spread of the MF
  void set_spread(double _spread) { spread = _spread; }
  double get_spread(void) { return spread; }

  // Set or return the type of the MF
  void set_type(int _type);
  int get_type(void) { return type; }
  
  // Evaluate the degree of membership for the given _input
  double find_dom(const double _input);

private:
  // MF type specific functions for evaluating the degree of membership
  inline double triangle(const double input);
  inline double gaussian(const double input);

  int type; // Type of the MF. Can be either TRIANGULAR_MF or EXPONENTIAL_MF.
  double center; // Center of the MF
  double spread; // Spread of the MF
};

typedef FUZZY_MF fuzzy_mf;

#endif // FUZZY_MF_HH
