/* 
 * FILE --- fuzzy_sys.hh
 *
 * PURPOSE --- Header file for a class for a multiple input multiple output 
 * (MIMO) standard fuzzy system. The user specifies the # of inputs and
 * outputs as well as the # of membership functions (MF) on each universe 
 * of discourse (UD). The MF can be defined as triangular with 50% of 
 * overlapp or as gaussian. Either minimum or product can be used for 
 * premises, and center of gravity (COG) or center average (CAV) for 
 * defuzzification.
 *
 * Uses two other utilities:
 *                  FUZZY_UD --- Fuzzy universe of discourse
 *                  FUZZY_MF --- Fuzzy membership function
 *
 */

// Prevent Multiple Inclusion
#ifndef FUZZY_SYS_HH
#define FUZZY_SYS_HH


// Include Files
#include <stdlib.h>
#include "fuzzy_ud.hh"


// Local Definitions and Macros
#define FUZZY_SYS_NO_ERROR 0
#define FUZZY_SYS_MEM_ERROR 1

#define DEFAULT_NUM_INPUTS 2 // Default # of input universes of discourse
#define DEFAULT_NUM_OUTPUTS 1 // Default # of output universes of discourse

#define DEFAULT_NUM_IN_MF 7 // Default # of MF on each input universe 
#define DEFAULT_NUM_OUT_MF 7 // Default # of MF on the output universe 

#define MINIMUM_PREMISE 0   // Use minimum rule for evaluating premises
#define PRODUCT_PREMISE 1   // Use product rule for evaluating premises

#define COG_DEFUZZ 0  // Use Center-of-gravity method for defuzzyfication
#define CAV_DEFUZZ 1  // Use Center-average method for defuzzyfication

#define DEFAULT_MF TRIANGULAR_MF  // Triangular MF is default
#define DEFAULT_PREMISE MINIMUM_PREMISE // Minimum is default for premise
#define DEFAULT_DEFUZZ COG_DEFUZZ  // COG defuzzification is default

#define MODE -1        // -1:"negative" feedback, 1:"positive" feedback
#define PI 3.14159265359

/* 
   Define the generic multiple input multiple output (MIMO) fuzzy controller
   class. 
*/
class FUZZY_SYS
{
public:
  
  /* Constructor:
     This is the constructor for the class. It allocates all the needed 
     memory and performs the required initialization for the fuzzy system.
     Parameters:
     _num_of_inputs = The # of inputs to the system
     _num_of_outputs = The # of outputs to the system
     _num_of_in_mfs = an array of size _num_of_inputs specifying
     the # of MF on each input UD. If NULL the # of input MF on all 
     input UD is set to DEFAULT_NUM_IN_MF
     _num_of_out_mfs = an array of size _num_of_outputs specifying
     the # of MF on each output UD. If NULL the # of output MF on all 
     output UD is set to DEFAULT_NUM_OUT_MF
  */
  FUZZY_SYS(int _num_of_inputs, 
	    int _num_of_outputs,
	    int *_num_of_in_mfs = NULL, 
	    int *_num_of_out_mfs = NULL);

  // Destructor
  ~FUZZY_SYS();

  /* Set the gains for each input UD
     _in_gains = an array of size equal to the # of inputs 
                 if NULL then all the gains are set to zero         
   */
  void set_input_gains(double *_in_gains = NULL);

  /* Set the gains for each output UD
     _out_gains = an array of size equal to the # of outputs 
                  if NULL then all the gains are set to zero         
   */
  void set_output_gains(double *_out_gains = NULL);

  /* Sets the rule for AND operationin the premise.
     _premise_type = The rule used for the AND operation in the premises
     of the rules. Can be either MINIMUM_PREMISE or PRODUCT_PREMISE. The 
     default is MINIMUM_PREMISE.
  */
  void set_premise_type(int _premise_type = MINIMUM_PREMISE);

   /* Sets the rule for defuzzification.
    _defuzz_type = The rule used for defuzzification. Can be either COG_DEFUZZ
     or CAV_DEFUZZ. The default is COG_DEFUZZ.
   */
  void set_defuzz_type(int _defuzz_type = COG_DEFUZZ);

  /* 
     The function that implements the controller: given an array of crisp 
     inputs evaluates the crisp outputs of the fuzzy controller

     input = an array of size equal to the # of inputs providing
             the current input values to the controller
     output = an array of size equal to the # of outputs which
     returns the corresponding outputs
   */
  void control(const double *input, double *output);

  // Performs a defuzzyfication and returns a crisp output.
  void defuzzyfy(double *output);

  // Get the private members in case we need it
  int get_premise_type(void) { return premise_type; };
  int get_defuzz_type(void) { return defuzz_type; };
  int get_num_of_inputs(void) { return num_of_inputs; };
  int get_num_of_outputs(void) { return num_of_outputs; };

  fuzzy_ud **in_ud;   // An array of input UD
  fuzzy_ud **out_ud;  // An array of output UD
  
  int fuzzy_sys_err; // Variable to hold the error condition

private:
  void allocate_memory(int *_num_of_in_mfs, int *_num_of_out_mfs);
  void free_memory(void);

  int num_of_inputs;  // # of inputs (to be specified by the user)
  int num_of_outputs;  // # of outputs (to be specified by the user)

  // Conveniance variables to be used during calculation of the 
  // controller output
  double *total_area;
  double *weighted_total_area;

  // Type of rule for evaluating premises. 
  // Can be either MINIMUM_PREMISE or PRODUCT_PREMISE.
  int premise_type; 

  // Type of defuzzyfication to be used.
  // Can be either COG_DEFUZZ or CAV_DEFUZZ.  
  int defuzz_type; 

  int num_of_rules; // # of the rules
};


#endif // FUZZY_SYS_HH
