/* 
 * FILE --- fuzzy_sys_ex.cc
 *
 * PURPOSE --- To provide an example use of the class for 4'th
 * order Runge-Kutta simulation, RK4_SIM, and the class for implementing
 * n-input m-output fuzzy system, FUZZY_SYS.
 *
 * Implements the main function of a class for performing a simulation of
 * a simple vehicle with point mass dynamics controlled by a fuzzy controller.
 * The fuzzy controller is implemented as a part of the class for the vehicle
 * dynamics - VEH_SIM_FUZZY. 
 *
 */

// Include Files
#include <stdlib.h>
#include <stdio.h>

#include "veh_sim_fuzzy.hh"
//#include "fuzzy_sys.hh"


void main(void) 
{
  double Ts = 0.01;             // Sampling time
  double y[3] = {-0.9*45.0,45.0,0.0};  // Initial conditions    
  double t0 = 0.0;              // Initial time 
  double tmax = 200.0;              // Total Simulation period
  
  // SISO fuzzy controller gains
  double input_gain = 1/15.0, output_gain = 2000.0;

  // # of simulation steps to be performed by the RK4 algorith
  // in one sampling interval
  int sim_per_cycle = 5;  

  // Create the vehicle object
  VEH_SIM_FUZZY my_vehicle(Ts,sim_per_cycle,t0,(double *)y);

  FILE *fp;  // To save data

  if (!(fp = fopen("veh_fuzzy_out.dat","w"))) {
    printf("Can't open file.\n");
    exit(1);
  }

  // Set the fuzzy controller gains
  my_vehicle.fuzzy_sys->set_input_gains(&input_gain);
  my_vehicle.fuzzy_sys->set_output_gains(&output_gain);

  /* Simulate For tmax seconds. */
  while(my_vehicle.current_time < tmax) 
    {
      my_vehicle.simulate();    
      fprintf(fp,"%f %f %f %f %f %f %f\n", my_vehicle.current_time,
	      my_vehicle.plant_state[0], my_vehicle.plant_state[1],
	      my_vehicle.plant_state[2], my_vehicle.ctrl_input[0],
	      my_vehicle.ctrl_input[1], my_vehicle.ctrl_input[2]);
    }

  fclose(fp);

  /* 
  // Simple example of use of the fuzzy system 
  double e[2], u[2];
  int in_mf[2] = {5, 5}, out_mf[2] = {5, 1};
  double input_gains[2] = {2.0/PI, 4.0/PI}, output_gains[2] = {20.0, 20.0};

  FUZZY_SYS my_fuzzy(2, 2, in_mf, out_mf);

  my_fuzzy.set_input_gains(input_gains);
  my_fuzzy.set_output_gains(output_gains);

  // Crips inputs.
  e[0] = -3.0;                 
  e[1] = 2.0*(PI/8.0 - PI/32.0);

  //  u = my_fuzzy.control(e);
  my_fuzzy.control(e, u);

  printf("e = %f, edot = %f, u1 = %f, u2 = %f\n",e[0],e[1],u[0],u[1]); 
*/

}



