/* 
 * FILE --- fuzzy_ud.cc
 *
 * PURPOSE --- Source file for a class for fuzzy universe of discourse (UD). 
 * The membership functions (MF) can be chosen to be normalized 
 * (i.e. b/n [-1, 1]) with equal spreads and 1/2 overlapp, or can be 
 * chosen b/n user defined range, [_first_center, _last_center] with user
 * defined equal spreads. In these options all the MF are the same user 
 * defined type. The possible MF types are triangular and Gaussian. 
 * Another option would be to initialize each individual MF with its own 
 * center, spread and type. 
 *
 * Uses theutility:
 *              FUZZY_MF --- Fuzzy membership function
 *
 */

// Include Files
#include "fuzzy_ud.hh"

// Define this macro if you want to use this utility without linking
// to the RCS Library
#ifdef NO_RCS
#include <stdio.h>            /* printf function */
#else
#include "rcs_prnt.hh"          /* rcs_print_ functions */
#endif


/*************************************************************/

FUZZY_UD::FUZZY_UD(int _num_of_mfs = DEFAULT_NUM_MF)
{
  // No error at start
  fuzzy_ud_err = FUZZY_UD_NO_ERROR;
  num_of_mfs = _num_of_mfs;

  if(num_of_mfs == 1) // We dont need this UD
    {
#ifdef NO_RCS
      printf("FUZZY_UD: Warning --- Only ONE MF on this UD!!!\n");
      printf("NOTE: You dont need this UD ...\n");
#else
      rcs_print("FUZZY_UD: Warning --- Only ONE MF on this UD!!!\n");
      rcs_print("NOTE: You dont need this UD ...\n");
#endif
    }

  // Allocate Memory
  allocate_memory();

  // If error then report
  if(fuzzy_ud_err != FUZZY_UD_NO_ERROR)
    {
#ifdef NO_RCS
      printf("FUZZY_UD: ERROR --- Not Enought Memory!!!\n");
      printf("The Fuzzy System will not function correctly ...\n");
#else
      rcs_print("FUZZY_UD: ERROR --- Not Enought Memory!!!\n");
      rcs_print("The Fuzzy System will not function correctly ...\n");
#endif
    }

  // Initialize centers and spreds as normalized 
  // triangular MF with 1/2 overlapp
  init_ud();

  // Initialize the gain for this UD to one
  ud_gain = 1.0;
}

FUZZY_UD::~FUZZY_UD()
{
  // Free Memory
  free_memory();
}

void FUZZY_UD::allocate_memory(void)
{
  // MF --- Triangular with centers at zero and spreads = 1.0
  if ((mfs = new fuzzy_mf[num_of_mfs]) == NULL)
    fuzzy_ud_err = FUZZY_UD_MEM_ERROR;

  // Degree of membership
  if ((deg_of_mbrship = new double[num_of_mfs]) == NULL)
    fuzzy_ud_err = FUZZY_UD_MEM_ERROR;
}

void FUZZY_UD::free_memory(void)
{
  delete[] mfs;
  delete[] deg_of_mbrship;
}

void FUZZY_UD::init_ud(int _mf_type = TRIANGULAR_MF, double _spread = 0)
{
  double step, spd, ctr;  // Conveniance variables

  // Initialize normalized centers and spreads
  if(num_of_mfs == 1) // We dont need this UD
    {
      mfs[0].set_type(_mf_type);
      mfs[0].set_spread(1.0);
      mfs[0].set_center(0.0);
    }
  else
    {
      // Step b/n each MF center
      step = 2.0/(num_of_mfs-1.0);

      // Choose the spread
      if(_spread <= 0) // Invalid spread or no spread specified
	{
	  spd = step;
	}
      else  // Valid spread specified
	{
	  spd = _spread;
	}

      // Start from -1.0 (the first center)
      ctr = -1.0;
      
      for (int ii=0; ii < num_of_mfs; ii++) 
	{
	  mfs[ii].set_type(_mf_type);
	  mfs[ii].set_spread(spd);
	  mfs[ii].set_center(ctr);
	  ctr += step;
	}
    }
}

void FUZZY_UD::init_ud(double _first_center, double _last_center,
		       int _mf_type, double _spread = 0)
{
  double step, spd, ctr;  // Conveniance variables

  // Init MF
  if(num_of_mfs == 1) // We dont need this UD
    {
      mfs[0].set_type(_mf_type);
      mfs[0].set_spread(1.0);
      mfs[0].set_center(0.0);
    }
  else
    {
      // Step b/n each MF center
      step = (_last_center - _first_center)/(num_of_mfs-1.0);
      
      // Choose the spread
      if(_spread <= 0) // Invalid spread or no spread specified
	{
	  spd = step;
	}
      else  // Valid spread specified
	{
	  spd = _spread;
	}
      
      // Start from the first center
      ctr = _first_center;
      
      // Initialize types, centers and spreads
      for (int ii=0; ii < num_of_mfs; ii++) 
	{
	  mfs[ii].set_type(_mf_type);
	  mfs[ii].set_spread(spd);
	  mfs[ii].set_center(ctr);
	  ctr += step;
	}
    }
}

void FUZZY_UD::init_ud(double *_centers, double *_spreads, int *_mf_types)
{
  // Initialize types, centers and spreads
  for (int ii=0; ii < num_of_mfs; ii++) 
    {
      mfs[ii].set_type(_mf_types[ii]);
      mfs[ii].set_spread(_spreads[ii]);
      mfs[ii].set_center(_centers[ii]);
    }
}

void FUZZY_UD::init_ud(fuzzy_mf *_mfs)
{
  // Initialize MF
  for (int ii=0; ii < num_of_mfs; ii++) 
    {
      mfs[ii].set_mf(_mfs[ii]);
    }
}

void FUZZY_UD::fuzzyfy(const double _input)
{
  scaled_input = SATURATE(ud_gain*_input, 1.0); 
  // Evaluate the degree of membership for each MF
  for (int ii=0; ii < num_of_mfs; ii++)
    {
      deg_of_mbrship[ii] = mfs[ii].find_dom(scaled_input);
    }
}

